/*
 * Program:	Convert mailbox formats
 *
 * Author:	Mark Crispin
 *		Networks and Distributed Computing
 *		Computing & Communications
 *		University of Washington
 *		Administration Building, AG-44
 *		Seattle, WA  98195
 *		Internet: MRC@CAC.Washington.EDU
 *
 * Date:	2 February 1994
 * Last Edited:	9 December 1997
 *
 * Copyright 1997 by the University of Washington
 *
 *  Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appears in all copies and that both the
 * above copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the University of Washington not be
 * used in advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  This software is made
 * available "as is", and
 * THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT LIMITATION ALL IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND IN
 * NO EVENT SHALL THE UNIVERSITY OF WASHINGTON BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT
 * (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <stdio.h>
#include <errno.h>
extern int errno;		/* just in case */
#include "mail.h"
#include "osdep.h"
#include "misc.h"
#include "linkage.h"
char *getpass ();		/* let's all do the cretin hop */

int critical = NIL;		/* flag saying in critical code */


/* Function prototypes */

void msg_string_init (STRING *s,void *data,unsigned long size);
char msg_string_next (STRING *s);
void msg_string_setpos (STRING *s,unsigned long i);
void main (int argc,char *argv[]);


/* Message pointer */

typedef struct msg_data {
  MAILSTREAM *stream;		/* stream */
  long msgno;			/* message number */
} MSGDATA;


/* Message string driver for message stringstructs */

STRINGDRIVER msg_string = {
  msg_string_init,		/* initialize string structure */
  msg_string_next,		/* get next byte in string structure */
  msg_string_setpos		/* set position in string structure */
};

/* Initialize file string structure for file stringstruct
 * Accepts: string structure
 *	    pointer to message data structure
 *	    size of string
 */

void msg_string_init (STRING *s,void *data,unsigned long size)
{
  MSGDATA *md = (MSGDATA *) data;
  s->data = data;		/* note stream/msgno and header length */
  mail_fetchheader_full (md->stream,md->msgno,NIL,&s->data1,FT_PREFETCHTEXT);
#if 0
  s->size = size;		/* message size */
#else	/* This kludge is necessary because of broken IMAP servers (sigh!) */
  mail_fetchtext_full (md->stream,md->msgno,&s->size,NIL);
  s->size += s->data1;		/* header + body size */
#endif
  SETPOS (s,0);
}


/* Get next character from file stringstruct
 * Accepts: string structure
 * Returns: character, string structure chunk refreshed
 */

char msg_string_next (STRING *s)
{
  char c = *s->curpos++;	/* get next byte */
  SETPOS (s,GETPOS (s));	/* move to next chunk */
  return c;			/* return the byte */
}


/* Set string pointer position for file stringstruct
 * Accepts: string structure
 *	    new position
 */

void msg_string_setpos (STRING *s,unsigned long i)
{
  MSGDATA *md = (MSGDATA *) s->data;
  if (i < s->data1) {		/* want header? */
    s->chunk = mail_fetchheader (md->stream,md->msgno);
    s->chunksize = s->data1;	/* header length */
    s->offset = 0;		/* offset is start of message */
  }
  else if (i < s->size) {	/* want body */
    s->chunk = mail_fetchtext (md->stream,md->msgno);
    s->chunksize = s->size - s->data1;
    s->offset = s->data1;	/* offset is end of header */
  }
  else {			/* off end of message */
    s->chunk = NIL;		/* make sure that we crack on this then */
    s->chunksize = 1;		/* make sure SNX cracks the right way... */
    s->offset = i;
  }
				/* initial position and size */
  s->curpos = s->chunk + (i -= s->offset);
  s->cursize = s->chunksize - i;
}

/* Main program */

void main (int argc,char *argv[])
{
  long i;
  unsigned long u;
  char *t,*t1,tmp[MAILTMPLEN],msgdate[MAILTMPLEN];
  DRIVER *d;
  STRING st;
  MESSAGECACHE *elt;
  MAILSTREAM *s = NIL;
#include "linkage.c"
  if (argc != 4) {		/* make sure correct number of arguments */
    puts ("usage: mbxcvt source_mailbox newtype destination_mailbox");
    _exit (1);
  }
  for (d = (DRIVER *) mail_parameters (NIL,GET_DRIVERS,NIL);
       d && strcmp (d->name,argv[2]); d = d->next);
  if (!d) {			/* make sure we found driver to use */
    printf ("?Unknown driver name %s\n",argv[2]);
    _exit (1);
  }
				/* open source mailbox */
  if (s = mail_open (s,argv[1],OP_READONLY)) {
    printf ("Mailbox %s is in %s format and has %ld message(s).\n",
	    s->mailbox,s->dtb->name,s->nmsgs);
				/* create destination mailbox */
    if (!(*d->create) (NIL,argv[3])) {
      printf ("?Can't create mailbox %s\n",argv[3]);
      mail_close (s);
      _exit (1);
    }
    if (s->nmsgs) {
      sprintf (tmp,"1:%ld",s->nmsgs);
      mail_fetchfast (s,tmp);	/* make sure we have all messages */
				/* copy messages */
      for (i = 1; i <= s->nmsgs; i++) {
	MSGDATA md;
				/* initialize flag string */
	memset (t = tmp,0,MAILTMPLEN);
				/* output system flags */
	if ((elt = mail_elt (s,i))->seen) strcat (t," \\Seen");
	if (elt->deleted) strcat (t," \\Deleted");
	if (elt->flagged) strcat (t," \\Flagged");
	if (elt->answered) strcat (t," \\Answered");
	if (elt->draft) strcat (t," \\Draft");
	if (u = elt->user_flags) do/* any user flags? */
	  if ((MAILTMPLEN - ((t += strlen (t)) - tmp)) >
	      (2 + strlen (t1 = s->user_flags[find_rightmost_bit (&u)]))) {
	    *t++ = ' ';		/* space delimiter */
	    strcpy (t,t1);	/* copy the user flag */
	  }
	while (u);		/* until no more user flags */
	md.stream = s;		/* init message data */
	md.msgno = i;		/* message number */
	INIT (&st,msg_string,(void *) &md,elt->rfc822_size);
	if (!(*d->append) (NIL,argv[3],tmp+1,mail_date (msgdate,elt),&st)) {
	  printf ("?Can't append message %ld to mailbox %s\n",i,argv[3]);
	  mail_close (s);
	  _exit (1);
	}
	mail_gc (s,GC_TEXTS);
      }
    }
    puts ("[Ok]");		/* make sure user knows it won */
    mail_close (s);		/* close off mailbox */
  }
}

/* Co-routines from MAIL library */


/* Message matches a search
 * Accepts: MAIL stream
 *	    message number
 */

void mm_searched (MAILSTREAM *stream,unsigned long msgno)
{
  fatal ("impossible mm_searched() call");
}


/* Message exists (i.e. there are that many messages in the mailbox)
 * Accepts: MAIL stream
 *	    message number
 */

void mm_exists (MAILSTREAM *stream,unsigned long number)
{
				/* dummy routine */
}


/* Message expunged
 * Accepts: MAIL stream
 *	    message number
 */

void mm_expunged (MAILSTREAM *stream,unsigned long number)
{
  fatal ("impossible mm_expunged() call");
}


/* Message flags update seen
 * Accepts: MAIL stream
 *	    message number
 */

void mm_flags (MAILSTREAM *stream,unsigned long number)
{
				/* dummy routine */
}

/* Mailbox found
 * Accepts: MAIL stream
 *	    hierarchy delimiter
 *	    mailbox name
 *	    mailbox attributes
 */

void mm_list (MAILSTREAM *stream,int delimiter,char *name,long attributes)
{
				/* dummy routine */
}


/* Subscribe mailbox found
 * Accepts: MAIL stream
 *	    hierarchy delimiter
 *	    mailbox name
 *	    mailbox attributes
 */

void mm_lsub (MAILSTREAM *stream,int delimiter,char *name,long attributes)
{
				/* dummy routine */
}


/* Mailbox status
 * Accepts: MAIL stream
 *	    mailbox name
 *	    mailbox status
 */

void mm_status (MAILSTREAM *stream,char *mailbox,MAILSTATUS *status)
{
				/* dummy routine */
}

/* Notification event
 * Accepts: MAIL stream
 *	    string to log
 *	    error flag
 */

void mm_notify (MAILSTREAM *stream,char *string,long errflg)
{
  mm_log (string,errflg);	/* just do mm_log action */
}


/* Log an event for the user to see
 * Accepts: string to log
 *	    error flag
 */

void mm_log (char *string,long errflg)
{
  switch (errflg) {  
  case BYE:
  case NIL:			/* no error */
    fprintf (stderr,"[%s]\n",string);
    break;
  case PARSE:			/* parsing problem */
  case WARN:			/* warning */
    fprintf (stderr,"%%%s\n",string);
    break;
  case ERROR:			/* error */
  default:
    fprintf (stderr,"?%s\n",string);
    break;
  }
}


/* Log an event to debugging telemetry
 * Accepts: string to log
 */

void mm_dlog (char *string)
{
  fprintf (stderr,"%s\n",string);
}

/* Get user name and password for this host
 * Accepts: parse of network mailbox name
 *	    where to return user name
 *	    where to return password
 *	    trial count
 */

void mm_login (NETMBX *mb,char *username,char *password,long trial)
{
  if (*mb->user) strcpy (username,mb->user);
  else {
    printf ("{%s/%s} username: ",mb->host,mb->service);
    gets (username);
  }
  strcpy (password,getpass ("password: "));
}


/* About to enter critical code
 * Accepts: stream
 */

void mm_critical (MAILSTREAM *stream)
{
  critical = T;			/* note in critical code */
}


/* About to exit critical code
 * Accepts: stream
 */

void mm_nocritical (MAILSTREAM *stream)
{
  critical = NIL;		/* note not in critical code */
}


/* Disk error found
 * Accepts: stream
 *	    system error code
 *	    flag indicating that mailbox may be clobbered
 * Returns: T if user wants to abort
 */

long mm_diskerror (MAILSTREAM *stream,long errcode,long serious)
{
  return T;
}


/* Log a fatal error event
 * Accepts: string to log
 */

void mm_fatal (char *string)
{
  fprintf (stderr,"?%s\n",string);
}
